"""
Tagulous test: Form TagField

Modules tested:
    tagulous.models.fields.TagField
    tagulous.forms.TagField
"""
from __future__ import absolute_import
from __future__ import unicode_literals

import django
from django import forms
from django.utils import six

from tests.lib import *


###############################################################################
####### Test form TagField
###############################################################################

class FormTagFieldTest(TagTestManager, TestCase):
    """
    Test form TagFields
    """
    def test_required(self):
        "Test required status is passed from field to widget"
        self.assertTrue(tag_forms.TagField(required=True).required)
        self.assertTrue(tag_forms.TagField(required=True).widget.is_required)
        self.assertFalse(tag_forms.TagField(required=False).required)
        self.assertFalse(tag_forms.TagField(required=False).widget.is_required)
        self.assertTrue(tag_forms.TagField().required)
        self.assertTrue(tag_forms.TagField().widget.is_required)

    def test_input(self):
        "Test valid and invalid input"
        self.assertFieldOutput(
            tag_forms.TagField,
            valid={
                'red': ['red'],
                'Red, Blue': ['Blue', 'Red'],
                '"red, blue", yellow': ['red, blue', 'yellow'],
            },
            invalid={},
            empty_value=[],
        )

    def test_force_lowercase(self):
        "Test force_lowercase flag set in constructor"
        self.assertFieldOutput(
            tag_forms.TagField,
            field_kwargs={
                'tag_options': tag_models.TagOptions(
                    force_lowercase=True
                )
            },
            valid={
                'red': ['red'],
                'Red, Blue': ['blue', 'red'],
                '"Red, Blue", Yellow': ['red, blue', 'yellow'],
            },
            invalid={},
            empty_value=[],
        )

    def test_response(self):
        "Test response is as expected"
        form = test_forms.TagFieldForm(data={'tags': 'red, blue'})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['tags'], ['blue', 'red'])

    def test_render_basics(self):
        "Check widget renders default settings (not required)"
        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField()
        form = LocalTestForm()
        self.assertHTMLEqual(six.text_type(form['tag']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tag" name="tag" {{required}}type="text" />'
        ))

    def test_render_tag_optional(self):
        "Check widget renders correctly when field not required"
        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(required=False)
        form = LocalTestForm()
        self.assertHTMLEqual(six.text_type(form['tag']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: false}" '
            'data-tagulous="true" '
            'id="id_tag" name="tag" type="text" />'
        ))

    def test_render_tag_list(self):
        "Check widget renders data-tag-list"
        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(
                autocomplete_tags=['one', 'two', 'three'],
            )
        form = LocalTestForm()
        self.assertHTMLEqual(six.text_type(form['tag']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'data-tag-list="'
            '[&quot;one&quot;, &quot;two&quot;, &quot;three&quot;]" '
            'id="id_tag" name="tag" {{required}}type="text" />'
        ))

    def test_render_tag_url(self):
        "Check widget renders data-tag-url"
        autocomplete_view = 'tagulous_tests_app-unlimited'
        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(
                tag_options=tag_models.TagOptions(
                    autocomplete_view=autocomplete_view,
                ),
            )
        form = LocalTestForm()
        self.assertHTMLEqual(six.text_type(form['tag']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'data-tag-url="'
            '/tagulous_tests_app/autocomplete/unlimited/" '
            'id="id_tag" name="tag" {{required}}type="text" />'
        ))

    def test_render_value(self):
        "Check widget renders value"
        form = test_forms.TagFieldForm(data={'tags': 'run, walk'})
        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tags" name="tags" {{required}}type="text" '
            'value="run, walk" />'
        ))

    def test_render_invalid_tag_url(self):
        "Check widget renders data-tag-url"
        autocomplete_view = 'tagulous_tests_app-view_does_not_exist'
        class LocalTestForm(forms.Form):
            tag = tag_forms.TagField(
                tag_options=tag_models.TagOptions(
                    autocomplete_view=autocomplete_view,
                ),
            )
        form = LocalTestForm()
        with self.assertRaises(ValueError) as cm:
            rendered = six.text_type(form['tag'])
        if django.VERSION < (1, 11):
            self.assertTrue(six.text_type(cm.exception).startswith(
                "Invalid autocomplete view: Reverse for '%s' with arguments '("
                ")' and keyword arguments '{}' not found." % autocomplete_view
            ))
        else:
            # Django 1.11 changes the error message
            self.assertTrue(six.text_type(cm.exception).startswith(
                "Invalid autocomplete view: Reverse for '{0}' not found. "
                "'{0}' is not a valid view function or pattern name.".format(
                    autocomplete_view,
                )
            ))

    def test_render_autocomplete_settings(self):
        "Check widget merges autocomplete settings with defaults"
        # Make a form with some autocomplete settings
        class LocalTestForm(forms.Form):
            tags = tag_forms.TagField(
                tag_options=tag_models.TagOptions(
                    autocomplete_settings={
                        'cats': 'purr',
                        'cows': 'moo',
                    },
                ),
            )
        form = LocalTestForm()

        # Set some defaults in the widget
        self.assertEqual(
            form['tags'].field.widget.default_autocomplete_settings, None
        )
        form['tags'].field.widget.default_autocomplete_settings = {
            'bees': 'buzz',
            'cats': 'mew',
        }

        # Render
        # Expecting bees:buzz, cats:purr, cows:moo
        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-options="{'
            '&quot;autocomplete_settings&quot;: {&quot;cows&quot;: &quot;moo'
            '&quot;, &quot;bees&quot;: &quot;buzz&quot;, &quot;cats&quot;: '
            '&quot;purr&quot;}, &quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tags" name="tags" {{required}}type="text" />'
        ))

    def test_invalid_prepare_value(self):
        "Check form field raises exception when given an invalid value"
        form = test_forms.TagFieldForm()
        with self.assertRaises(ValueError) as cm:
            form['tags'].field.prepare_value(['tag', 'list', 'fails'])
        self.assertEqual(
            six.text_type(cm.exception),
            "Tag field could not prepare unexpected value",
        )


###############################################################################
####### Test ModelForm TagField
###############################################################################

class ModelFormTagFieldTest(TagTestManager, TestCase):
    """
    Test form TagFields
    """
    manage_models = [
        test_models.TagFieldModel,
    ]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.form = test_forms.TagFieldModelForm
        self.model = test_models.TagFieldModel
        self.tag_model = self.model.tags.tag_model

    def test_formfield(self):
        "Test that model.TagField.formfield works correctly"
        # Check that the model fields are generated correctly
        tag1_field = self.model._meta.get_field('tags').formfield()
        self.assertIsInstance(tag1_field, tag_forms.TagField)
        self.assertIsInstance(tag1_field.tag_options, tag_models.TagOptions)

        # Check we can get it from the descriptor shortcut method
        tag2_field = self.model.tags.formfield()
        self.assertIsInstance(tag2_field, tag_forms.TagField)
        self.assertIsInstance(tag2_field.tag_options, tag_models.TagOptions)

    def test_media(self):
        "Test that a model form with a TagField has the correct media"
        media = self.form().media
        for js in tag_settings.AUTOCOMPLETE_JS:
            self.assertTrue(js in media._js)
        for grp, files in tag_settings.AUTOCOMPLETE_CSS.items():
            self.assertTrue(grp in media._css)
            for css in files:
                self.assertTrue(css in media._css[grp])

    def test_model_form_save(self):
        """
        Test that a model form with a TagField saves correctly
        """
        form = test_forms.TagFieldModelForm(
            data={
                'name': 'Test 1',
                'tags': 'blue, red',
            }
        )
        self.assertTrue(form.is_valid())
        t1 = form.save()

        # Check in-memory instance
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.tags, 'blue, red')

        # Check database
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue': 1,
            'red': 1,
        })

    def test_model_form_save_commit_false(self):
        """
        Test that a model form with a TagField saves correctly when
        commit=False
        """
        form = test_forms.TagFieldModelForm(
            data={
                'name': 'Test 1',
                'tags': 'blue, red',
            }
        )
        self.assertTrue(form.is_valid())
        t1 = form.save(commit=False)
        t1.save()

        # Check in-memory instance
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.tags, '')
        self.assertTagModel(self.tag_model, {})
        self.assertInstanceEqual(t1, name='Test 1', tags='')

        # Save M2M data
        form.save_m2m()

        # Check in-memory instance
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.tags, 'blue, red')

        # Check database
        self.assertInstanceEqual(t1, name='Test 1', tags='blue, red')
        self.assertTagModel(self.tag_model, {
            'blue': 1,
            'red': 1,
        })

    def test_override_option_dict_formfield(self):
        "Test defaults can be overriden in formfield() call with a dict"
        # Confirm default
        field1 = self.form().fields['tags']
        self.assertEqual(field1.tag_options.case_sensitive, False)

        # Change default
        field2 = self.model.tags.formfield(
            tag_options={'case_sensitive': True}
        )
        self.assertEqual(field2.tag_options.case_sensitive, True)

    def test_override_option_cls_formfield(self):
        "Test defaults can be overriden in formfield() call with a TagOption"
        # Confirm default
        field1 = self.form().fields['tags']
        self.assertEqual(field1.tag_options.case_sensitive, False)

        # Change default
        field2 = self.model.tags.formfield(
            tag_options=tag_models.TagOptions(case_sensitive=True)
        )
        self.assertEqual(field2.tag_options.case_sensitive, True)

    def test_override_autocomplete_tags_formfield(self):
        "Test list of autocomplete tags can be passed in formfield"
        self.tag_model.objects.create(name='red')
        self.tag_model.objects.create(name='blue')
        self.tag_model.objects.create(name='green')

        # Confirm default
        field1 = self.form().fields['tags']
        self.assertSequenceEqual(
            [t.name for t in field1.autocomplete_tags],
            [t.name for t in self.tag_model.objects.all()],
        )

        # Change default
        field2 = self.model.tags.formfield(
            autocomplete_tags=['pink', 'lime']
        )
        self.assertSequenceEqual(field2.autocomplete_tags, ['pink', 'lime'])

    def test_render_tag_list(self):
        "Check widget renders data-tag-list"
        self.tag_model.objects.create(name='red')
        self.tag_model.objects.create(name='blue')
        self.tag_model.objects.create(name='yellow')
        self.assertTagModel(self.tag_model, {
            'red': 0,
            'blue': 0,
            'yellow': 0,
        })
        form = self.form(data={'name': 'Test 1', 'tags': 'red, blue'})
        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'data-tag-list="'
            '[&quot;blue&quot;, &quot;red&quot;, &quot;yellow&quot;]" '
            'id="id_tags" name="tags" {{required}}type="text" '
            'value="red, blue" />'
        ))

    def test_initial_string(self):
        "Check initial tag string"
        form = test_forms.TagFieldForm(initial={'tags': 'red, blue'})
        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tags" name="tags" {{required}}type="text" '
            'value="red, blue" />'
        ))

    def test_initial_tag_list(self):
        "Check initial tag list"
        t1 = self.tag_model.objects.create(name='red')
        t2 = self.tag_model.objects.create(name='blue')
        form = test_forms.TagFieldForm(initial={'tags': [t1, t2]})
        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tags" name="tags" {{required}}type="text" '
            'value="blue, red" />'
        ))

    def test_initial_tag_queryset(self):
        "Check initial tag queryset"
        self.tag_model.objects.create(name='red')
        self.tag_model.objects.create(name='blue')
        tags = self.tag_model.objects.all()
        form = test_forms.TagFieldForm(initial={'tags': tags})
        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tags" name="tags" {{required}}type="text" '
            'value="blue, red" />'
        ))

    def test_tagged_edit(self):
        "Check edit tagged model form instance works"
        t1 = self.model.objects.create(name='Test 1', tags='blue, red')

        form = self.form(instance=t1)

        self.assertHTMLEqual(six.text_type(form['tags']), (
            '<input autocomplete="off" '
            'data-tag-list="[&quot;blue&quot;, &quot;red&quot;]" '
            'data-tag-options="{&quot;required&quot;: true}" '
            'data-tagulous="true" '
            'id="id_tags" name="tags" {{required}}type="text" '
            'value="blue, red" />'
        ))

    def test_tagmeta_without_autocomplete_settings(self):
        """
        Check that a tag widget copes with a tag field which takes its options
        from a TagModel with a TagMeta, but where the TagMeta is missing
        autocomplete_settings, and the widget has default_autocomplete_settings.
        """
        class TagMetaUserForm(forms.ModelForm):
            class Meta:
                model = test_models.TagMetaUser
                exclude = []
        form = TagMetaUserForm()
        form['two'].field.widget.default_autocomplete_settings = {
            'bees': 'buzz',
        }

        self.assertHTMLEqual(six.text_type(form['two']), (
            '<input autocomplete="off" '
            'data-tag-list="[]" '
            'data-tag-options="{&quot;autocomplete_settings&quot;: '
            '{&quot;bees&quot;: &quot;buzz&quot;}, '
            '&quot;case_sensitive&quot;: true, '
            '&quot;force_lowercase&quot;: true, &quot;max_count&quot;: 10, '
            '&quot;required&quot;: false}" '
            'data-tagulous="true" '
            'id="id_two" name="two" type="text" />'
        ))

###############################################################################
#######  Test TagField blank
###############################################################################

class ModelFormTagFieldOptionalTest(TagTestManager, TestCase):
    "Test optional TagField"
    manage_models = [
        test_models.TagFieldOptionalModel,
    ]

    def setUpExtra(self):
        self.form = test_forms.TagFieldOptionalModelForm
        self.model = test_models.TagFieldOptionalModel
        self.tag_model = self.model.tag.tag_model

    def test_optional_tagfield(self):
        "Check model with optional tagfield can be saved when empty"
        form = self.form(data={'name': 'Test 1'})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.tag.count(), 0)
        self.assertTagModel(self.tag_model, {})


class ModelFormTagFieldRequiredTest(TagTestManager, TestCase):
    "Test required TagField"
    manage_models = [
        test_models.TagFieldRequiredModel,
    ]

    def setUpExtra(self):
        self.form = test_forms.TagFieldRequiredModelForm
        self.model = test_models.TagFieldRequiredModel
        self.tag_model = self.model.tag.tag_model

    def test_required_tagfield_set(self):
        "Check model with required tagfield can be saved when set"
        form = self.form(data={'name': 'Test 1', 'tag':  'red, blue'})
        self.assertTrue(form.is_valid())
        t1 = form.save()
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.tag, 'blue, red')
        self.assertEqual(six.text_type(t1.tag), 'blue, red')
        self.assertIsInstance(t1.tag, tag_models.TagRelatedManagerMixin)
        self.assertEqual(t1.tag.count(), 2)
        self.assertEqual(t1.tag.all()[0], 'blue')
        self.assertEqual(t1.tag.all()[1], 'red')
        self.assertTagModel(self.tag_model, {
            'red': 1,
            'blue': 1,
        })

    def test_required_tagfield_empty(self):
        "Check model with required tagfield cannot be saved when empty"
        form = self.form(data={'name': 'Test 1'})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], 'tag')
        self.assertEqual(len(form.errors['tag']), 1)
        self.assertEqual(form.errors['tag'][0], 'This field is required.')


###############################################################################
####### Test form TagField options
###############################################################################

class ModelFormTagFieldOptionsTest(TagTestManager, TestCase):
    manage_models = [
        test_models.TagFieldOptionsModel,
    ]

    def setUpExtra(self):
        # Load initial tags for all models which have them
        self.form = test_forms.TagFieldOptionsModelForm
        self.model = test_models.TagFieldOptionsModel

    @skip_if_mysql
    def test_case_sensitive_true(self):
        "Test form TagField case_sensitive true"
        # Prep tag model
        tag_model = self.model.case_sensitive_true.tag_model
        self.assertTagModel(tag_model, {
            'Adam': 0,
        })

        # Check that the option is passed to the form field from model field
        form = self.form(data={'case_sensitive_true': 'adam'})
        field = form.fields['case_sensitive_true']
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.case_sensitive, True)

        # Check that the option is passed to the widget
        self.assertEqual(field.widget.tag_options.case_sensitive, True)

        # No effect on form data
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['case_sensitive_true'], ['adam'])

        # No effect on save
        obj = form.save()
        self.assertEqual(obj.case_sensitive_true, 'adam')
        self.assertTagModel(tag_model, {
            'Adam': 0,
            'adam': 1,
        })

    @skip_if_mysql
    def test_case_sensitive_false(self):
        "Test form TagField case_sensitive false"
        # Prep tag model
        tag_model = self.model.case_sensitive_false.tag_model

        # Check that the option is passed to the form field from model field
        form = self.form(data={'case_sensitive_false': 'adam'})
        field = form.fields['case_sensitive_false']
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.case_sensitive, False)
        self.assertEqual(field.widget.tag_options.case_sensitive, False)

        # Still no effect on form data
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['case_sensitive_false'], ['adam'])

        # But when saved, the instance will be capitalised
        obj = form.save()
        self.assertEqual(obj.case_sensitive_false, 'Adam')
        self.assertTagModel(tag_model, {
            'Adam': 1,
        })

    def test_force_lowercase_true(self):
        "Test form TagField force_lowercase true"
        # Prep tag model
        tag_model = self.model.force_lowercase_true.tag_model

        # Check that the option is passed to the form field from model field
        form = self.form(data={'force_lowercase_true': 'Adam'})
        field = form.fields['force_lowercase_true']
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.force_lowercase, True)
        self.assertEqual(field.widget.tag_options.force_lowercase, True)

        # Form data returns lowercase case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['force_lowercase_true'], ['adam'])

        # When saved, the instance will be lowercase
        obj = form.save()
        self.assertEqual(obj.force_lowercase_true, 'adam')
        self.assertTagModel(tag_model, {
            'adam': 1,
        })

    def test_force_lowercase_false(self):
        "Test form TagField force_lowercase false"
        # Check that the option is passed to the form field from model field
        form = self.form(data={'force_lowercase_false': 'Adam'})
        field = form.fields['force_lowercase_false']
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.force_lowercase, False)
        self.assertEqual(field.widget.tag_options.force_lowercase, False)

        # Check the input string is returned in the new case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['force_lowercase_false'], ['Adam'])

    def test_max_count(self):
        "Test form TagField max_count is passed and enforced"
        # It's not relevant
        # Check that the option is passed to the form field from model field
        form = self.form(data={'max_count': 'one'})
        field = form.fields['max_count']
        self.assertIsInstance(field, tag_forms.TagField)
        self.assertEqual(field.tag_options.max_count, 3)

        # Check the input string is returned in the new case
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['max_count'], ['one'])

        # Check two
        form = self.form(data={'max_count': 'one, two'})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['max_count'], ['one', 'two'])

        # Check three
        form = self.form(data={'max_count': 'one, two, three'})
        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data['max_count'], sorted(['one', 'two', 'three'])
        )

        # Check four
        form = self.form(data={'max_count': 'one, two, three, four'})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], 'max_count')
        self.assertEqual(len(form.errors['max_count']), 1)
        self.assertEqual(
            form.errors['max_count'][0],
            'This field can only have 3 arguments',
        )

    def text_max_count_1(self):
        "Test form TagField max_count of 1"
        # Mostly just to test grammar of the error message
        class LocalTestForm(forms.Form):
            tags = tag_forms.TagField(
                tag_options=tag_models.TagOptions(
                    max_count=1,
                ),
            )
        form = self.form(data={'max_count': 'one'})
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['max_count'], ['one'])

        # Check two
        form = self.form(data={'max_count': 'one, two'})
        self.assertFalse(form.is_valid())
        self.assertEqual(len(form.errors.keys()), 1)
        self.assertEqual(list(form.errors.keys())[0], 'max_count')
        self.assertEqual(len(form.errors['max_count']), 1)
        self.assertEqual(
            form.errors['max_count'][0],
            'This field can only have 1 argument',
        )

    def test_initial_without_autocomplete_initial(self):
        "Check a field with initial but without autocomplete_initial lists all"
        tag_model = self.model.initial_string.tag_model
        tag_model.objects.create(name='David')
        tag_model.objects.create(name='Eric')
        tag_model.objects.create(name='Frank')

        # Confirm default
        field = self.form().fields['initial_string']
        self.assertSequenceEqual(
            [t.name for t in field.autocomplete_tags],
            [t.name for t in tag_model.objects.all()],
        )

    def test_initial_with_autocomplete_initial(self):
        "Check a field with initial and autocomplete_initial lists initial"
        tag_model = self.model.initial_list.tag_model
        tag_model.objects.create(name='David')
        tag_model.objects.create(name='Eric')
        tag_model.objects.create(name='Frank')

        # Confirm default
        field = self.form().fields['initial_list']
        self.assertSequenceEqual(
            [t.name for t in field.autocomplete_tags],
            [t.name for t in tag_model.objects.initial()],
        )



"""
Tagulous models

Make important functions and classes available on tagulous.models
"""
from tagulous.models.options import TagOptions
from tagulous.models.models import (
    BaseTagModel, TagModel, BaseTagTreeModel, TagTreeModel,
    TagModelQuerySet, TagModelManager,
)
from tagulous.models.managers import (
    SingleTagManager, TagRelatedManagerMixin,
)
from tagulous.models.descriptors import (
    BaseTagDescriptor, SingleTagDescriptor, TagDescriptor,
)
from tagulous.models.fields import (
    BaseTagField, SingleTagField, TagField,
    singletagfields_from_model, tagfields_from_model,
)

from tagulous.models import initial
from tagulous.models import migrations
from tagulous.models.tagged import TaggedModel, TaggedManager, TaggedQuerySet



"""
Tagulous test: Forms with SingleTagField and TagFields

Modules tested:
    tagulous.models.fields.SingleTagField
    tagulous.models.fields.TagField
    tagulous.forms.SingleTagField
    tagulous.forms.TagField
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six

from tests.lib import *


###############################################################################
####### Test mixed form
###############################################################################

class FormMixedNonTagRefTest(TagTestManager, TestCase):
    """
    Test form TagFields
    """
    manage_models = [
        test_models.MixedNonTagRefTest,
    ]

    def setUpExtra(self):
        self.model = test_models.MixedNonTagRefTest
        self.tag_model = self.model.singletag.tag_model

    def test_model_form_save(self):
        """
        Test that a model form with a TagField saves correctly with other
        relationships
        """
        tag1 = self.tag_model.objects.create(name='blue')
        tag2 = self.tag_model.objects.create(name='red')
        tag3 = self.tag_model.objects.create(name='green')
        form = test_forms.MixedNonTagRefModelForm(
            data={
                'name': 'Test 1',
                'singletag': 'purple',
                'tags': 'yellow, orange',
                'fk': tag1.pk,
                'mm': [tag2.pk, tag3.pk],
            }
        )
        self.assertTrue(form.is_valid())
        t1 = form.save()

        # Check in-memory instance
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.singletag, 'purple')
        self.assertEqual(t1.tags, 'orange, yellow')
        self.assertEqual(t1.fk, tag1)
        self.assertEqual(list(t1.mm.all().order_by('pk')), [tag2, tag3])

        # Check database
        self.assertInstanceEqual(
            t1, name='Test 1', singletag='purple', tags='orange, yellow',
            fk=tag1, mm=[tag2, tag3],
        )
        self.assertTagModel(self.tag_model, {
            'blue': 0,
            'red': 0,
            'green': 0,
            'purple': 1,
            'orange': 1,
            'yellow': 1,
        })

    def test_model_form_save_commit_false(self):
        """
        Test that a model form with a TagField saves correctly when save_m2m
        is also called
        """
        tag1 = self.tag_model.objects.create(name='blue')
        tag2 = self.tag_model.objects.create(name='red')
        tag3 = self.tag_model.objects.create(name='green')
        form = test_forms.MixedNonTagRefModelForm(
            data={
                'name': 'Test 1',
                'singletag': 'purple',
                'tags': 'orange, yellow',
                'fk': tag1.pk,
                'mm': [tag2.pk, tag3.pk],
            }
        )
        self.assertTrue(form.is_valid())
        t1 = form.save(commit=False)
        t1.save()
        form.save_m2m()

        # Check in-memory instance
        self.assertEqual(t1.name, 'Test 1')
        self.assertEqual(t1.singletag, 'purple')
        self.assertEqual(t1.tags, 'orange, yellow')
        self.assertEqual(t1.fk, tag1)
        self.assertEqual(list(t1.mm.all().order_by('pk')), [tag2, tag3])

        # Check database
        self.assertInstanceEqual(
            t1, name='Test 1', singletag='purple', tags='orange, yellow',
            fk=tag1, mm=[tag2, tag3],
        )
        self.assertTagModel(self.tag_model, {
            'blue': 0,
            'red': 0,
            'green': 0,
            'purple': 1,
            'orange': 1,
            'yellow': 1,
        })


###############################################################################
####### Test inline formsets
###############################################################################

class SingleInlineFormsetTest(TagTestManager, TestCase):
    """
    Test inline formset for SingleTagField relationships
    """
    manage_models = [
        test_models.SimpleMixedTest,
    ]

    def setUpExtra(self):
        self.model = test_models.SimpleMixedTest
        self.tag_model = self.model.singletag.tag_model
        self.formset = test_forms.SimpleMixedSingleInline

    def test_add(self):
        tag1 = self.tag_model.objects.create(name='Mr')
        self.assertTagModel(self.tag_model, {'Mr': 0})
        self.assertEqual(self.model.objects.count(), 0)

        formset = self.formset(
            data={
                'formset-TOTAL_FORMS': '3',
                'formset-INITIAL_FORMS': '0',
                'formset-MAX_NUM_FORMS': '1000',

                'formset-0-name': 'Test 1',
                'formset-0-DELETE': '',

                'formset-1-name': 'Test 2',
                'formset-1-DELETE': '',
            },
            instance=tag1,
            prefix='formset'
        )
        self.assertTrue(formset.is_valid())
        formset.save()

        self.assertEqual(self.model.objects.count(), 2)
        obj1 = self.model.objects.get(name='Test 1')
        obj2 = self.model.objects.get(name='Test 2')
        self.assertEqual(obj1.singletag, tag1)
        self.assertEqual(obj2.singletag, tag1)
        self.assertTagModel(self.tag_model, {'Mr': 2})

    def test_edit_pk(self):
        "Check that edit formset forms refer to the tag pk, not its name"
        obj1 = self.model.objects.create(name='Test 1', singletag='Mr')
        obj2 = self.model.objects.create(name='Test 2', singletag='Mr')
        tag1 = obj1.singletag
        self.assertTagModel(self.tag_model, {'Mr': 2})
        self.assertEqual(self.model.objects.count(), 2)

        formset = self.formset(prefix='formset', instance=tag1)
        self.assertHTMLEqual(six.text_type(formset.forms[0]['singletag']), (
            '<input type="hidden" name="formset-0-singletag" '
            'value="%d" id="id_formset-0-singletag" />'
        ) % tag1.pk)


    def test_edit(self):
        obj1 = self.model.objects.create(name='Test 1', singletag='Mr')
        obj2 = self.model.objects.create(name='Test 2', singletag='Mr')
        tag1 = obj1.singletag
        self.assertTagModel(self.tag_model, {'Mr': 2})
        self.assertEqual(self.model.objects.count(), 2)

        formset = self.formset(
            data={
                'formset-TOTAL_FORMS': '5',
                'formset-INITIAL_FORMS': '2',
                'formset-MAX_NUM_FORMS': '1000',

                'formset-0-name': 'Test 1e',
                'formset-0-id': '%d' % obj1.pk,
                'formset-0-singletag': '%d' % tag1.pk,
                'formset-0-DELETE': '',

                'formset-1-name': 'Test 2e',
                'formset-1-id': '%d' % obj2.pk,
                'formset-1-singletag': '%d' % tag1.pk,
                'formset-1-DELETE': '',
            },
            instance=tag1,
            prefix='formset'
        )
        formset.is_valid()
        self.assertTrue(formset.is_valid())
        formset.save()

        self.assertEqual(self.model.objects.count(), 2)
        obj1 = self.model.objects.get(name='Test 1e')
        obj2 = self.model.objects.get(name='Test 2e')
        self.assertEqual(obj1.singletag, tag1)
        self.assertEqual(obj2.singletag, tag1)
        self.assertTagModel(self.tag_model, {'Mr': 2})



"""
Test models
"""
from __future__ import unicode_literals

from django.db import models
from django.utils import six

try:
    import django.apps as django_apps
except ImportError:
    django_apps = None

import tagulous


#
# MigrationTestModel will be created as needed
#

app_name = 'tagulous_tests_migration'


def clear_django():
    "Clear app from django's model cache"
    # Models to clear out from loading cache
    model_names = [
        # Test model
        'migrationtestmodel',

        # Tagulous models
        'tagulous_migrationtestmodel_singletag',
        'tagulous_migrationtestmodel_tags',

        # Django through models
        'migrationtestmodel_tags',
    ]

    if django_apps:
        # Django 1.7 or later
        # Clear the get_models LRU cache
        django_apps.apps.clear_cache()

        # Find dict container for app models cache
        app_config = django_apps.apps.get_app_config(app_name)
        app_models = app_config.models

    else:
        # Django 1.6 or earlier
        # Clear loading cache so they'll be reloaded for next get_models call
        models.loading.cache._get_models_cache.clear()

        # Find dict container for app models cache
        app_models = models.loading.cache.app_models.get(app_name, {})

    # Clear the named models from the app models cache
    for lower_name in model_names:
        try:
            del app_models[lower_name]
        except KeyError:
            pass

def unset_model():
    "Remove the model"
    if 'MigrationTestModel' in globals():
        del globals()['MigrationTestModel']
    clear_django()

def set_model_initial():
    "Create an initial model without tag fields"
    clear_django()
    return type(
        str("MigrationTestModel"), (models.Model,), {
            '__module__': 'tests.tagulous_tests_migration.models',
            'name': models.CharField(max_length=10),
        }
    )

def set_model_tagged():
    "Initial model with tag fields"
    clear_django()
    model = type(
        str("MigrationTestModel"), (models.Model,), {
            '__module__': 'tests.tagulous_tests_migration.models',
            'name': models.CharField(max_length=10),
            'singletag': tagulous.models.SingleTagField(blank=True, null=True),
            # Django 1.4 hates unicode intermediary table name
            str('tags'): tagulous.models.TagField(),
        }
    )

    # Just confirm dynamic creation worked as expected
    assert issubclass(model, tagulous.models.tagged.TaggedModel), 'Model is not tagged'
    assert issubclass(model.singletag.tag_model, tagulous.models.models.TagModel), 'Single tag model not TagModel'
    assert issubclass(model.tags.tag_model, tagulous.models.models.TagModel), 'Tag model not TagModel'
    return model

def set_model_tree():
    "Tagged model with tags field as tree"
    clear_django()
    model = type(
        str("MigrationTestModel"), (models.Model,), {
            '__module__': 'tests.tagulous_tests_migration.models',
            'name': models.CharField(max_length=10),
            'singletag': tagulous.models.SingleTagField(blank=True, null=True),
            str('tags'): tagulous.models.TagField(tree=True),
        }
    )

    # Just confirm dynamic creation worked as expected
    assert issubclass(model, tagulous.models.tagged.TaggedModel), 'Model is not tagged'
    assert issubclass(model.singletag.tag_model, tagulous.models.models.TagModel), 'Single tag model not TagModel'
    assert issubclass(model.tags.tag_model, tagulous.models.models.TagModel), 'Tag model not TagModel'
    return model



import os
import re
import sys
from setuptools import setup, find_packages

VERSION = "0.14.1"

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def runtests(args):
    "Run tests"
    import django
    from django.conf import settings
    from django.core.management import execute_from_command_line

    if not settings.configured:
        testenv = re.sub(
            r'[^a-zA-Z0-9]', '_',
            os.environ.get('TOXENV', '_'.join(str(v) for v in django.VERSION)),
        )
        tests_migration_module_name = 'migrations_{}'.format(testenv)

        SETTINGS = dict(
            INSTALLED_APPS=[
                'django.contrib.auth',
                'django.contrib.admin',
                'django.contrib.sessions',
                'django.contrib.contenttypes',
                'django.contrib.messages',
                'tagulous',
                'tests',
                'tests.tagulous_tests_app',
                'tests.tagulous_tests_app2',
                'tests.tagulous_tests_migration',
            ],
            MIDDLEWARE=[
                'django.middleware.common.CommonMiddleware',
                'django.contrib.sessions.middleware.SessionMiddleware',
                'django.contrib.auth.middleware.AuthenticationMiddleware',
                'django.contrib.messages.middleware.MessageMiddleware',
            ],
            ROOT_URLCONF='tests.tagulous_tests_app.urls',
            SERIALIZATION_MODULES={
                'xml': 'tagulous.serializers.xml_serializer',
                'json': 'tagulous.serializers.json',
                'python': 'tagulous.serializers.python',
            },
            TEMPLATES=[{
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'APP_DIRS': True,
                'OPTIONS': {
                    'context_processors': [
                        'django.contrib.auth.context_processors.auth',
                        'django.contrib.messages.context_processors.messages',
                    ],
                },
            }],
            MIGRATION_MODULES={
                'tagulous_tests_migration': 'tests.tagulous_tests_migration.{}'.format(tests_migration_module_name),
            },
            TAGULOUS_NAME_MAX_LENGTH=191,
        )
        SETTINGS['SOUTH_MIGRATION_MODULES'] = SETTINGS['MIGRATION_MODULES']

        # If yaml is available, add to serialisers
        try:
            import yaml
        except ImportError:
            pass
        else:
            SETTINGS['SERIALIZATION_MODULES']['yaml'] = 'tagulous.serializers.pyyaml'

        # If south is required and available, add to installed apps
        if django.VERSION < (1, 7):
            try:
                import south
            except ImportError:
                import warnings
                warnings.warn(
                    "south not installed, needed for tests for Django %s" % (
                        '.'.join(str(v) for v in django.VERSION)
                    ),
                    ImportWarning
                )
            else:
                SETTINGS['INSTALLED_APPS'] += ['south']

        # Backwards compatibility for middleware
        if django.VERSION < (1, 10):
            SETTINGS['MIDDLEWARE_CLASSES'] = SETTINGS['MIDDLEWARE']

        # Build database settings
        DATABASE = {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': ':memory:',
        }
        engine = os.environ.get('DATABASE_ENGINE')
        if engine:
            if engine == "pgsql":
                DATABASE['ENGINE'] = 'django.db.backends.postgresql_psycopg2'
                DATABASE['HOST'] = 'localhost'
            elif engine == "mysql":
                DATABASE['ENGINE'] = 'django.db.backends.mysql'

                # Make sure test DB is going to be UTF8
                if django.VERSION < (1, 7):
                    DATABASE_TEST = {
                        'TEST_CHARSET': 'utf8',
                        'TEST_COLLATION': 'utf8_general_ci',
                    }
                else:
                    DATABASE_TEST = {
                        'TEST': {
                            'CHARSET': 'utf8',
                            'COLLATION': 'utf8_general_ci',
                        }
                    }
                DATABASE.update(DATABASE_TEST)

            else:
                raise ValueError("Unknown database engine")

            DATABASE['NAME'] = os.environ.get('DATABASE_NAME', 'test_tagulous_%s' % testenv)
            for key in ['USER', 'PASSWORD', 'HOST', 'PORT']:
                if 'DATABASE_' + key in os.environ:
                    DATABASE[key] = os.environ['DATABASE_' + key]
        SETTINGS['DATABASES'] = {'default': DATABASE, 'test': DATABASE}

        # Make sure the django migration loader will find MIGRATION_MODULES
        # This will be cleaned away after each test
        tests_migration_path = os.path.join(
            os.path.dirname(__file__), 'tests', 'tagulous_tests_migration',
        )
        if not os.path.isdir(tests_migration_path):
            raise ValueError('tests.tagulous_tests_migration not found')

        tests_migration_module_path = os.path.join(
            tests_migration_path, tests_migration_module_name,
        )
        if not os.path.exists(tests_migration_module_path):
            os.mkdir(tests_migration_module_path)

        tests_migration_module_init = os.path.join(
            tests_migration_module_path, '__init__.py',
        )
        if not os.path.exists(tests_migration_module_init):
            open(tests_migration_module_init, 'a').close()

        # Configure
        settings.configure(**SETTINGS)

    execute_from_command_line(args[:1] + ['test'] + (args[2:] or ['tests']))


if len(sys.argv) > 1 and sys.argv[1] == 'test':
    runtests(sys.argv)
    sys.exit()

setup(
    name = "django-tagulous",
    version = VERSION,
    author = "Richard Terry",
    author_email = "code@radiac.net",
    description = ("Fabulous Tagging for Django"),
    license = "BSD",
    keywords = "django tagging",
    url = "http://radiac.net/projects/django-tagulous/",
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 4 - Beta',
        #'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Django',
        'Framework :: Django :: 1.4',
        'Framework :: Django :: 1.5',
        'Framework :: Django :: 1.6',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
    ],
    extras_require = {
        'dev': ['tox', 'jasmine'],
        'devdb': ['psycopg2', 'mysqlclient'],
        'i18n': ['unidecode'],
    },
    zip_safe=True,
    packages=find_packages(exclude=('docs', 'tests*',)),
    include_package_data=True,
)



"""
Backwards and forwards compatibility for template tags
"""
from __future__ import unicode_literals

import django
from django import template
from django.utils import six

register = template.Library()


#
# {% url "quoted.view" %}
#
# Django 1.4 provides quote support with {% load url from future %}
# Django 1.7 deprecates future
# Django 1.9 removes future
#

if django.VERSION < (1, 5):
    from django.templatetags.future import url as django_url
else:
    from django.template.defaulttags import url as django_url

@register.tag
def url(parser, token):
    """
    Quoted urls, introduced in Django 1.5, available in 1.4+
    """
    return django_url(parser, token)



"""
JSON serializer with Tagulous support
"""
from __future__ import unicode_literals

from django.core.serializers import json as json_serializer
from django.utils import six

from tagulous.serializers import base


class Serializer(base.SerializerMixin, json_serializer.Serializer):
    """
    JSON serializer with tag field support
    """
    pass


Deserializer = base.DeserializerWrapper(
    json_serializer.Deserializer,
    doc="Deserialize a stream or string of JSON data, with tag field support",
)



# -*- coding: utf-8 -*-
"""
Tagulous test: Tag models

Modules tested:
    tagulous.models.models.BaseTagModel
    tagulous.models.models.TagModel
    tagulous.models.models.TagModelManager
    tagulous.models.models.TagModelQuerySet
"""
from __future__ import absolute_import
from __future__ import unicode_literals

from django.utils import six
import tagulous.settings as tagulous_settings

from tagulous.settings import SLUG_TRUNCATE_UNIQUE
from tests.lib import *


class TagModelTest(TagTestManager, TestCase):
    """
    Test tag model basics
    """
    manage_models = [
        test_models.MixedTest,
        test_models.MixedRefTest,
        test_models.NonTagRefTest,
    ]

    def setUpExtra(self):
        self.tag_model = test_models.MixedTestTagModel
        self.model1 = test_models.MixedTest
        self.model2 = test_models.MixedRefTest
        self.model_nontag = test_models.NonTagRefTest

    def test_tags_equal_instance(self):
        "Test TagModel.__eq__ with instances"
        t1a = self.tag_model.objects.create(name='one')
        t1b = self.tag_model.objects.get(name='one')
        self.assertEqual(t1a, t1b)

    def test_tags_equal_string(self):
        "Test TagModel.__eq__ with instance and string"
        t1 = self.tag_model.objects.create(name='one')
        self.assertEqual(t1, 'one')

    def test_tags_not_equal_instance(self):
        "Test TagModel.__ne__ with instances"
        t1 = self.tag_model.objects.create(name='one')
        t2 = self.tag_model.objects.create(name='two')
        self.assertNotEqual(t1, t2)

    def test_tags_not_equal_string(self):
        "Test TagModel.__eq__ with instance and string"
        t1 = self.tag_model.objects.create(name='one')
        self.assertNotEqual(t1, 'two')

    def test_get_absolute_url_defined(self):
        "Test get_absolute_url when passed in field definition"
        t1 = self.tag_model.objects.create(name='one')
        self.assertEqual(t1.get_absolute_url(), 'url for one')

    def test_get_absolute_url_not_defined(self):
        "Test get_absolute_url when passed in field definition"
        t1 = test_models.SimpleMixedTest.tags.tag_model.objects.create(name='one')
        with self.assertRaises(AttributeError) as cm:
            t1.get_absolute_url()
        self.assertEqual(
            six.text_type(cm.exception),
            "'Tagulous_SimpleMixedTest_tags' has no attribute 'get_absolute_url'"
        )

    def test_name_field_length(self):
        """
        Value is initialized in setup.py runtests() settings
        """
        self.assertEqual(self.tag_model._meta.get_field('name').max_length,
                         tagulous_settings.NAME_MAX_LENGTH)

    def test_name_field_length(self):
        """
        Value is initialized in setup.py runtests() settings
        """
        self.assertEqual(self.tag_model._meta.get_field('slug').max_length,
                         tagulous_settings.SLUG_MAX_LENGTH)

    def assertRelatedExists(self, related_fields, match_model, field_name):
        """
        Look through the related fields and find the field which refers to the
        specified model; fail if it does not exist
        """
        match_field = match_model._meta.get_field(field_name)
        for related in related_fields:
            if django.VERSION < (1, 8):
                rel_model = related.model
            else:
                rel_model = related.related_model

            if rel_model == match_model and related.field == match_field:
                return related
        self.fail('Expected related field not found')

    def test_get_related_fields(self):
        "Check the class method returns a list of related fields"
        related_fields = self.tag_model.get_related_fields()
        self.assertEqual(len(related_fields), 4)

        # FK comes before M2M, so first two will be the SingleTagFields
        self.assertRelatedExists(related_fields, self.model1, 'singletag')
        self.assertRelatedExists(related_fields, self.model2, 'singletag')

        # Now the TagFields
        self.assertRelatedExists(related_fields, self.model1, 'tags')
        self.assertRelatedExists(related_fields, self.model2, 'tags')

    def test_get_related_fields_standard(self):
        "Check the class method can also find standard relationships"
        related_fields = self.tag_model.get_related_fields(include_standard=True)
        self.assertEqual(len(related_fields), 6)

        # SingleTagFields/FKs
        self.assertRelatedExists(related_fields, self.model1, 'singletag')
        self.assertRelatedExists(related_fields, self.model2, 'singletag')
        self.assertRelatedExists(related_fields, self.model_nontag, 'fk')

        # TagFields/M2Ms
        self.assertRelatedExists(related_fields, self.model1, 'tags')
        self.assertRelatedExists(related_fields, self.model2, 'tags')
        self.assertRelatedExists(related_fields, self.model_nontag, 'mm')

    def test_get_related_objects(self):
        """
        Check the class method returns a list of related models, fields and
        instances
        """
        t1 = self.create(self.model1, name="Test 1", singletag='Mr', tags='blue')
        t2 = self.create(self.model1, name="Test 2", singletag='Mr', tags='green')
        t3 = self.create(self.model2, name="Test 3", singletag='Mrs', tags='green')
        t4 = self.create(self.model2, name="Test 4", singletag='Mr', tags='green')

        #
        # Check Mr
        #
        singletag1 = self.tag_model.objects.get(name='Mr')
        rel_st1 = singletag1.get_related_objects()
        self.assertEqual(len(rel_st1), 2)
        # Don't assume order - we could probably guess reliably, but
        # why risk it in tests
        # rel_st1 is list of [model, field, [instances]]
        if rel_st1[0][0] == self.model1:
            rel_st1_m1 = rel_st1[0]
            rel_st1_m2 = rel_st1[1]
        else:
            rel_st1_m1 = rel_st1[1]
            rel_st1_m2 = rel_st1[0]

        self.assertEqual(rel_st1_m1[0], self.model1)
        self.assertEqual(rel_st1_m1[1], self.model1._meta.get_field('singletag'))
        self.assertEqual(len(rel_st1_m1[2]), 2)
        self.assertEqual(rel_st1_m1[2][0], t1)
        self.assertEqual(rel_st1_m1[2][1], t2)

        self.assertEqual(rel_st1_m2[0], self.model2)
        self.assertEqual(rel_st1_m2[1], self.model2._meta.get_field('singletag'))
        self.assertEqual(len(rel_st1_m2[2]), 1)
        self.assertEqual(rel_st1_m2[2][0], t4)

        #
        # Check Mrs
        #
        singletag2 = self.tag_model.objects.get(name='Mrs')
        rel_st2 = singletag2.get_related_objects()
        self.assertEqual(len(rel_st2), 1)
        rel_st2_m1 = rel_st2[0]
        self.assertEqual(rel_st2_m1[0], self.model2)
        self.assertEqual(rel_st2_m1[1], self.model2._meta.get_field('singletag'))
        self.assertEqual(len(rel_st2_m1[2]), 1)
        self.assertEqual(rel_st2_m1[2][0], t3)

        #
        # Check blue
        #
        tags1 = self.tag_model.objects.get(name='blue')
        rel_t1 = tags1.get_related_objects()
        self.assertEqual(len(rel_t1), 1)
        rel_t1_m1 = rel_t1[0]
        self.assertEqual(rel_t1_m1[0], self.model1)
        self.assertEqual(rel_t1_m1[1], self.model1._meta.get_field('tags'))
        self.assertEqual(len(rel_t1_m1[2]), 1)
        self.assertEqual(rel_t1_m1[2][0], t1)

        #
        # Check green
        #
        tags2 = self.tag_model.objects.get(name='green')
        rel_t2 = tags2.get_related_objects()
        self.assertEqual(len(rel_t2), 2)
        if rel_t2[0][0] == self.model1:
            rel_t2_m1 = rel_t2[0]
            rel_t2_m2 = rel_t2[1]
        else:
            rel_t2_m1 = rel_t2[1]
            rel_t2_m2 = rel_t2[0]

        self.assertEqual(rel_t2_m1[0], self.model1)
        self.assertEqual(rel_t2_m1[1], self.model1._meta.get_field('tags'))
        self.assertEqual(len(rel_t2_m1[2]), 1)
        self.assertEqual(rel_t2_m1[2][0], t2)

        self.assertEqual(rel_t2_m2[0], self.model2)
        self.assertEqual(rel_t2_m2[1], self.model2._meta.get_field('tags'))
        self.assertEqual(len(rel_t2_m2[2]), 2)
        self.assertEqual(rel_t2_m2[2][0], t3)
        self.assertEqual(rel_t2_m2[2][1], t4)

    def test_get_related_objects_flat(self):
        "Check the class method returns a flat list of related instances"
        t1 = self.create(self.model1, name="Test 1", singletag='Mr', tags='blue')
        t2 = self.create(self.model1, name="Test 2", singletag='Mr', tags='green')
        t3 = self.create(self.model2, name="Test 3", singletag='Mrs', tags='green')
        t4 = self.create(self.model2, name="Test 4", singletag='Mr', tags='green')

        # Check Mr
        singletag1 = self.tag_model.objects.get(name='Mr')
        rel_st1 = singletag1.get_related_objects(flat=True)
        self.assertEqual(len(rel_st1), 3)
        rel_st1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_st1[0], t1)
        self.assertEqual(rel_st1[1], t2)
        self.assertEqual(rel_st1[2], t4)

        # Check Mrs
        singletag2 = self.tag_model.objects.get(name='Mrs')
        rel_st2 = singletag2.get_related_objects(flat=True)
        self.assertEqual(len(rel_st2), 1)
        self.assertEqual(rel_st2[0], t3)

        # Check blue
        tags1 = self.tag_model.objects.get(name='blue')
        rel_t1 = tags1.get_related_objects(flat=True)
        self.assertEqual(len(rel_t1), 1)
        self.assertEqual(rel_t1[0], t1)

        # Check green
        tags2 = self.tag_model.objects.get(name='green')
        rel_t2 = tags2.get_related_objects(flat=True)
        self.assertEqual(len(rel_t2), 3)
        rel_t2.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t2[0], t2)
        self.assertEqual(rel_t2[1], t3)
        self.assertEqual(rel_t2[2], t4)

    def test_get_related_objects_flat_distinct(self):
        "Check the class method returns a flat list of distinct related instances"
        t1 = self.create(self.model1, name="Test 1", singletag='Mr', tags='blue')
        t2 = self.create(self.model2, name="Test 2", singletag='blue', tags='blue')

        # Check blue non-distinct
        tags1 = self.tag_model.objects.get(name='blue')
        rel_t1 = tags1.get_related_objects(flat=True)
        self.assertEqual(len(rel_t1), 3)
        rel_t1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t1[0], t1)
        self.assertEqual(rel_t1[1], t2)
        self.assertEqual(rel_t1[2], t2)

        # Check blue distinct
        rel_t2 = tags1.get_related_objects(flat=True, distinct=True)
        self.assertEqual(len(rel_t2), 2)
        rel_t2.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t2[0], t1)
        self.assertEqual(rel_t2[1], t2)

    def test_get_related_objects_flat_include_standard(self):
        """
        Check the class method returns a flat list of related instances,
        including standard relationships

        No need to test other options with include_standard, uses same code
        """
        t1 = self.create(self.model1, name="Test 1", singletag='Mr', tags='blue')
        singletag1 = self.tag_model.objects.get(name='Mr')
        tags1 = self.tag_model.objects.get(name='blue')
        t2 = self.create(self.model_nontag, name="Test 2", fk=singletag1, mm=[tags1])

        # Check Mr
        rel_st1 = singletag1.get_related_objects(flat=True, include_standard=True)
        self.assertEqual(len(rel_st1), 2)
        rel_st1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_st1[0], t1)
        self.assertEqual(rel_st1[1], t2)

        # Check blue
        rel_t1 = tags1.get_related_objects(flat=True, include_standard=True)
        self.assertEqual(len(rel_t1), 2)
        rel_t1.sort(key=lambda tag: tag.name)
        self.assertEqual(rel_t1[0], t1)
        self.assertEqual(rel_t1[1], t2)

    def test_increment(self):
        "Increment the tag count"
        tag1 = self.create(self.tag_model, name='blue')
        self.assertInstanceEqual(tag1, count=0)
        tag1.increment()
        self.assertInstanceEqual(tag1, count=1)

    def test_increment_db(self):
        "Increment the tag count using the DB value, not in-memory"
        tag1 = self.create(self.tag_model, name='blue')
        tag2 = self.tag_model.objects.get(pk=tag1.pk)
        self.assertEqual(tag1.count, 0)
        self.assertEqual(tag2.count, 0)

        tag1.increment()
        self.assertInstanceEqual(tag1, count=1)
        self.assertInstanceEqual(tag2, count=1)
        self.assertEqual(tag1.count, 1)
        self.assertEqual(tag2.count, 0)

        tag2.increment()
        self.assertInstanceEqual(tag1, count=2)
        self.assertInstanceEqual(tag2, count=2)
        self.assertEqual(tag1.count, 1)
        self.assertEqual(tag2.count, 2)

    def test_decrement(self):
        tag1 = self.create(self.tag_model, name='blue', count=2)
        self.assertTagModel(self.tag_model, {
            'blue': 2,
        })
        tag1.decrement()
        self.assertTagModel(self.tag_model, {
            'blue': 1,
        })

    def test_decrement_delete(self):
        tag1 = self.create(self.tag_model, name='blue', count=1)
        self.assertTagModel(self.tag_model, {
            'blue': 1,
        })
        tag1.decrement()
        self.assertTagModel(self.tag_model, {})

    def test_decrement_delete_protected(self):
        tag1 = self.create(self.tag_model, name='blue', count=1, protected=True)
        self.assertTagModel(self.tag_model, {
            'blue': 1,
        })
        tag1.decrement()
        self.assertTagModel(self.tag_model, {
            'blue': 0,
        })

    def test_decrement_delete_hasrefs(self):
        """
        Check that when a tag's count hits 0, but still has non-tag field
        references, that it isn't deleted - don't want to cascade/break refs
        """
        # Create tags with false count
        tag1 = self.create(self.tag_model, name='blue', count=1)
        tag2 = self.create(self.tag_model, name='red', count=1)
        self.assertTagModel(self.tag_model, {
            'blue': 1,
            'red':  1,
        })

        # Create object with conventional references to tags
        t1 = self.create(
            self.model_nontag, name="CharField", fk=tag1, mm=[tag2]
        )
        self.assertInstanceEqual(
            t1, name="CharField", fk=tag1, mm=[tag2]
        )

        # No change to count
        self.assertTagModel(self.tag_model, {
            'blue': 1,
            'red':  1,
        })

        # Check get_related_objects knows about them
        self.assertEqual(
            len(tag1.get_related_objects(flat=True, include_standard=True)), 1
        )
        self.assertEqual(
            len(tag2.get_related_objects(flat=True, include_standard=True)), 1
        )

        # Now decrement counts to 0, but tags remain
        tag1.decrement()
        tag2.decrement()
        self.assertTagModel(self.tag_model, {
            'blue': 0,
            'red':  0,
        })

    def test_update_count(self):
        "Purposely knock the count off and update it"
        t1 = self.create(self.model1, name="Test 1", tags='blue')
        tag1 = self.tag_model.objects.get(name='blue')
        self.assertTagModel(self.tag_model, {'blue': 1})
        tag1.count = 3
        tag1.save()
        self.assertTagModel(self.tag_model, {'blue': 3})
        tag1.update_count()
        self.assertTagModel(self.tag_model, {'blue': 1})
        t1.delete()
        self.assertTagModel(self.tag_model, {})

    def test_slug_set(self):
        "Check the slug field is set correctly"
        t1a = self.tag_model.objects.create(name='One and Two!')
        self.assertEqual(t1a.slug, 'one-and-two')

    def test_slug_saved(self):
        "Check the slug field is saved correctly"
        t1a = self.tag_model.objects.create(name='One and Two!')
        t1b = self.tag_model.objects.get(name='One and Two!')
        self.assertEqual(t1a.slug, t1b.slug)
        self.assertEqual(t1b.slug, 'one-and-two')

    def test_slug_clash(self):
        "Check slug field avoids clashes to remain unique"
        t1a = self.tag_model.objects.create(name='one and two')
        t2a = self.tag_model.objects.create(name='One and Two!')
        t3a = self.tag_model.objects.create(name='One and Two?')
        t4a = self.tag_model.objects.create(name='One and Two.')
        self.assertEqual(t1a.slug, 'one-and-two')
        self.assertEqual(t2a.slug, 'one-and-two_1')
        self.assertEqual(t3a.slug, 'one-and-two_2')
        self.assertEqual(t4a.slug, 'one-and-two_3')

        t1b = self.tag_model.objects.get(name='one and two')
        t2b = self.tag_model.objects.get(name='One and Two!')
        t3b = self.tag_model.objects.get(name='One and Two?')
        t4b = self.tag_model.objects.get(name='One and Two.')
        self.assertEqual(t1b.slug, 'one-and-two')
        self.assertEqual(t2b.slug, 'one-and-two_1')
        self.assertEqual(t3b.slug, 'one-and-two_2')
        self.assertEqual(t4b.slug, 'one-and-two_3')

    def test_tag_model_factory(self):
        "Check the tag model factory supports setting max lengths"
        TestModel = test_models.TagSlugShorterModel
        self.assertEqual(TestModel._meta.get_field('name').max_length, 20)
        self.assertEqual(TestModel._meta.get_field('slug').max_length, 10)

    def test_long_slug_truncates(self):
        "Check the slug field is truncated correctly"
        TestModel = test_models.TagSlugShorterModel
        name_length = TestModel._meta.get_field('name').max_length
        slug_length = TestModel._meta.get_field('slug').max_length
        self.assertLess(slug_length, name_length)
        long_name = 'x' * name_length
        t1a = TestModel.objects.create(name='x' * name_length)
        self.assertEqual(t1a.name, 'x' * name_length)
        self.assertEqual(t1a.slug, 'x' * slug_length)

    def test_long_slug_clash_truncates(self):
        "Check clashing slug fields are truncated correctly"
        TestModel = test_models.TagSlugShorterModel
        name_length = TestModel._meta.get_field('name').max_length
        slug_length = TestModel._meta.get_field('slug').max_length
        self.assertLess(slug_length, name_length)
        long_name = 'x' * name_length
        ln1 = '{}{}'.format('x' * (name_length - 1), '1')
        ln2 = '{}{}'.format('x' * (name_length - 1), '2')
        t1a = TestModel.objects.create(name=ln1)
        t2a = TestModel.objects.create(name=ln2)
        self.assertEqual(t1a.name, ln1)
        self.assertEqual(t1a.slug, 'x' * slug_length)
        self.assertEqual(t2a.name, ln2)
        slug2 = '{}_{}'.format('x' * (slug_length - SLUG_TRUNCATE_UNIQUE), '1')
        self.assertEqual(t2a.slug, slug2)


###############################################################################
####### Test TagMeta in tag model
###############################################################################

class TagMetaTest(TagTestManager, TestCase):
    """
    Test TagMeta class. Builds on tests in tests_options.
    """
    def test_sets_options(self):
        "Check TagMeta sets the options"
        opt = test_models.TagMetaAbstractModel.tag_options

        # Check local options
        cls_opt = opt.items(with_defaults=False)
        self.assertEqual(cls_opt['initial'], ['Adam', 'Brian', 'Chris'])
        self.assertEqual(cls_opt['force_lowercase'], True)
        self.assertEqual(cls_opt['max_count'], 5)
        self.assertTrue('case_sensitive' not in cls_opt)

        # Check default options
        self.assertEqual(opt.case_sensitive, False)

    def test_inheritance(self):
        "Check TagMeta can be inherited and overridden"
        opt_abstract = test_models.TagMetaAbstractModel.tag_options
        opt = test_models.TagMetaModel.tag_options

        # Check they're not shared instances
        self.assertNotEqual(id(opt_abstract), id(opt))
        self.assertNotEqual(id(opt), id(tag_models.models.BaseTagModel.tag_options))

        # Check local options
        cls_opt = opt.items(with_defaults=False)
        self.assertEqual(cls_opt['case_sensitive'], True)
        self.assertEqual(cls_opt['max_count'], 10)

        # Local options will also include inherited options
        self.assertEqual(opt.initial, ['Adam', 'Brian', 'Chris'])
        self.assertEqual(opt.force_lowercase, True)
        self.assertEqual(opt.max_count, 10)
        self.assertEqual(opt.case_sensitive, True)


###############################################################################
####### Test unicode in tag model
###############################################################################

class TagModelUnicodeTest(TagTestManager, TestCase):
    """
    Test unicode tags - forced to not use unidecode, even if available
    """
    manage_models = [
        test_models.MixedTest,
    ]

    def setUpExtra(self):
        # Disable unidecode support
        self.unidecode_status = tag_utils.unidecode
        tag_utils.unidecode = None

        self.model = test_models.MixedTest
        self.tag_model = test_models.MixedTestTagModel
        self.o1 = self.create(
            self.model, name="Test",
            singletag='男の子',
            tags='boy, niño, 男の子',
        )

    def tearDownExtra(self):
        tag_utils.unidecode = self.unidecode_status

    def test_setup(self):
        "Check setup created tags as expected"
        self.assertTagModel(self.tag_model, {
            'boy':     1,
            'niño':    1,
            '男の子':   2,
        })


    # Check lookup

    def test_get_singletag_get(self):
        "Check unicode singletag name matches"
        t1 = self.model.objects.get(singletag='男の子')
        self.assertEqual(t1.pk, self.o1.pk)

    def test_get_tag_ascii(self):
        "Check unicode tag name matches when ascii"
        t1 = self.model.objects.get(tags='boy')
        self.assertEqual(t1.pk, self.o1.pk)

    def test_get_tag_extended_ascii(self):
        "Check unicode tag name matches when extended ascii"
        t1 = self.model.objects.get(tags='niño')
        self.assertEqual(t1.pk, self.o1.pk)

    def test_get_tag_japanese(self):
        "Check unicode tag name matches when above extended ascii"
        t1 = self.model.objects.get(tags='男の子')
        self.assertEqual(t1.pk, self.o1.pk)


    # Check render

    def test_singletag_render(self):
        "Check unicode singletag name renders"
        t1 = self.model.objects.get(name="Test")
        self.assertEqual(six.text_type(t1.singletag), '男の子')

    def test_tag_render(self):
        "Check unicode tag name renders"
        t1 = self.model.objects.get(name="Test")
        tags = list(t1.tags.all())
        self.assertEqual(six.text_type(tags[0]), 'boy')
        self.assertEqual(six.text_type(tags[1]), 'niño')
        self.assertEqual(six.text_type(tags[2]), '男の子')

    def test_tag_string_render(self):
        "Check unicode tags string renders"
        t1 = self.model.objects.get(name="Test")
        self.assertEqual(six.text_type(t1.tags), 'boy, niño, 男の子')


    # Check slugs

    def test_slug_ascii(self):
        "Check unicode tag name slugified when ascii"
        t1 = self.tag_model.objects.get(name="boy")
        self.assertEqual(t1.slug, 'boy')

    def test_slug_extended_ascii(self):
        "Check unicode tag name slugified when ascii"
        t1 = self.tag_model.objects.get(name="niño")
        self.assertEqual(t1.slug, 'nino')

    def test_slug_japanese(self):
        "Check unicode tag name slugified when ascii"
        t1 = self.tag_model.objects.get(name="男の子")
        self.assertEqual(t1.slug, '___')


try:
    from unidecode import unidecode
except ImportError:
    unidecode = None

@unittest.skipIf(unidecode is None, 'optional unidecode not installed')
class TagModelUnicodeUnidecodeTest(TagTestManager, TestCase):
    """
    Test unicode tags, without using unidecode

    This only affects slugs
    """
    manage_models = [
        test_models.MixedTest,
    ]

    def setUpExtra(self):
        self.model = test_models.MixedTest
        self.tag_model = test_models.MixedTestTagModel
        self.o1 = self.create(
            self.model, name="Test",
            singletag='男の子',
            tags='boy, niño, 男の子',
        )

    def test_setup(self):
        "Check setup created tags as expected"
        self.assertTagModel(self.tag_model, {
            'boy':     1,
            'niño':    1,
            '男の子':   2,
        })


    # unidecode only affects slugs

    def test_slug_ascii(self):
        "Check unicode tag name slugified when ascii"
        t1 = self.tag_model.objects.get(name="boy")
        self.assertEqual(t1.slug, 'boy')

    def test_slug_extended_ascii(self):
        "Check unicode tag name slugified when ascii"
        t1 = self.tag_model.objects.get(name="niño")
        self.assertEqual(t1.slug, 'nino')

    def test_slug_japanese(self):
        "Check unicode tag name slugified when ascii"
        t1 = self.tag_model.objects.get(name="男の子")
        self.assertEqual(t1.slug, 'nan-nozi')


###############################################################################
####### Test tag merging
###############################################################################

class TagModelMergeTest(TagTestManager, TestCase):
    """
    Test tag merging
    """
    manage_models = [
        test_models.MixedTest,
        test_models.MixedRefTest,
    ]

    def test_merge_tags(self):
        tag_model = test_models.MixedTestTagModel

        # Set up database
        a1 = self.create(test_models.MixedTest, name='a1', singletag='one', tags='one')
        a2 = self.create(test_models.MixedTest, name='a2', singletag='two', tags='two')
        a3 = self.create(test_models.MixedTest, name='a3', singletag='three', tags='three')

        b1 = self.create(test_models.MixedRefTest, name='b1', singletag='one', tags='one')
        b2 = self.create(test_models.MixedRefTest, name='b2', singletag='two', tags='two')
        b3 = self.create(test_models.MixedRefTest, name='b3', singletag='three', tags='three')

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'one': 4,
            'two': 4,
            'three': 4,
        })
        self.assertInstanceEqual(a1, singletag='one', tags='one')
        self.assertInstanceEqual(a2, singletag='two', tags='two')
        self.assertInstanceEqual(a3, singletag='three', tags='three')
        self.assertInstanceEqual(b1, singletag='one', tags='one')
        self.assertInstanceEqual(b2, singletag='two', tags='two')
        self.assertInstanceEqual(b3, singletag='three', tags='three')

        # Merge tags
        s1 = tag_model.objects.get(name='one')
        s1.merge_tags(
            tag_model.objects.filter(name__in=['one', 'two', 'three'])
        )

        # Check it's correct
        self.assertTagModel(tag_model, {
            'one': 12,
        })
        self.assertInstanceEqual(a1, singletag='one', tags='one')
        self.assertInstanceEqual(a2, singletag='one', tags='one')
        self.assertInstanceEqual(a3, singletag='one', tags='one')
        self.assertInstanceEqual(b1, singletag='one', tags='one')
        self.assertInstanceEqual(b2, singletag='one', tags='one')
        self.assertInstanceEqual(b3, singletag='one', tags='one')

    def test_merge_multiple_tags(self):
        "Test merging a queryset of multiple tags"
        tag_model = test_models.MixedTestTagModel

        # Set up database
        t1 = self.create(test_models.MixedTest, name='Test 1', tags='blue, green, red')
        t2 = self.create(test_models.MixedTest, name='Test 2', tags='blue, green, red')

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'blue': 2,
            'green': 2,
            'red': 2,
        })
        self.assertInstanceEqual(t1, tags='blue, green, red')
        self.assertInstanceEqual(t2, tags='blue, green, red')

        # Merge tags
        s1 = tag_model.objects.get(name='blue')
        s1.merge_tags(
            tag_model.objects.filter(name__in=['blue', 'green', 'red'])
        )

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'blue': 2,
        })
        self.assertInstanceEqual(t1, tags='blue')
        self.assertInstanceEqual(t2, tags='blue')

    def test_merge_by_name(self):
        "Test merging a list of tag names, including tags which don't exist"
        tag_model = test_models.MixedTestTagModel

        # Set up database
        t1 = self.create(test_models.MixedTest, name='Test 1', tags='blue, green, red')
        t2 = self.create(test_models.MixedTest, name='Test 2', tags='blue, green, red')

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'blue': 2,
            'green': 2,
            'red': 2,
        })
        self.assertInstanceEqual(t1, tags='blue, green, red')
        self.assertInstanceEqual(t2, tags='blue, green, red')

        # Merge tags
        s1 = tag_model.objects.get(name='blue')
        s1.merge_tags(['blue', 'green', 'red', 'pink'])

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'blue': 2,
        })
        self.assertInstanceEqual(t1, tags='blue')
        self.assertInstanceEqual(t2, tags='blue')

    def test_merge_by_obj_list(self):
        "Test merging a list of tag objects"
        tag_model = test_models.MixedTestTagModel
        t1 = self.create(test_models.MixedTest, name='Test 1', tags='blue, green, red')
        t2 = self.create(test_models.MixedTest, name='Test 2', tags='blue, green, red')

        # Merge tags
        s1 = tag_model.objects.get(name='blue')
        s1.merge_tags(list(tag_model.objects.all()))

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'blue': 2,
        })
        self.assertInstanceEqual(t1, tags='blue')
        self.assertInstanceEqual(t2, tags='blue')

    def test_merge_by_tag_string(self):
        "Test merging a tag string, including tags which don't exist"
        tag_model = test_models.MixedTestTagModel
        t1 = self.create(test_models.MixedTest, name='Test 1', tags='blue, green, red')
        t2 = self.create(test_models.MixedTest, name='Test 2', tags='blue, green, red')

        # Merge tags
        s1 = tag_model.objects.get(name='blue')
        s1.merge_tags('blue, green, red, pink')

        # Confirm it's correct
        self.assertTagModel(tag_model, {
            'blue': 2,
        })
        self.assertInstanceEqual(t1, tags='blue')
        self.assertInstanceEqual(t2, tags='blue')


###############################################################################
####### Test tag model manager and queryset
###############################################################################

class TagModelQuerySetTest(TagTestManager, TestCase):
    """
    Test tag model queryset and manager
    """
    manage_models = [
        test_models.TagFieldOptionsModel,
    ]

    def setUpExtra(self):
        self.model = test_models.TagFieldOptionsModel
        self.tag_model = self.model.initial_list.tag_model
        self.o1 = self.model.objects.create(
            name='Test 1',
            initial_list='David, Eric',
        )
        self.o2 = self.model.objects.create(
            name='Test 2',
            initial_list='Eric, Frank',
        )

    def test_setup(self):
        self.assertTagModel(self.model.initial_list, {
            'Adam':  0,
            'Brian': 0,
            'Chris': 0,
            'David': 1,
            'Eric':  2,
            'Frank': 1,
        })

    def test_initial(self):
        initial_only = self.tag_model.objects.initial()
        self.assertEqual(len(initial_only), 3)
        self.assertEqual(initial_only[0], 'Adam')
        self.assertEqual(initial_only[1], 'Brian')
        self.assertEqual(initial_only[2], 'Chris')

    def test_filter_or_initial(self):
        filtered = self.tag_model.objects.filter_or_initial(
            tagfieldoptionsmodel__name='Test 1',
        )
        self.assertEqual(len(filtered), 5)
        self.assertEqual(filtered[0], 'Adam')
        self.assertEqual(filtered[1], 'Brian')
        self.assertEqual(filtered[2], 'Chris')
        self.assertEqual(filtered[3], 'David')
        self.assertEqual(filtered[4], 'Eric')

    def test_weight_scale_up(self):
        "Test weight() scales up to max"
        # Scale them to 2+2n: 0=2, 1=4, 2=6
        weighted = self.tag_model.objects.weight(min=2, max=6)
        self.assertEqual(len(weighted), 6)
        self.assertEqual(weighted[0].name, 'Adam')
        self.assertEqual(weighted[0].weight, 2)

        self.assertEqual(weighted[1], 'Brian')
        self.assertEqual(weighted[1].weight, 2)

        self.assertEqual(weighted[2], 'Chris')
        self.assertEqual(weighted[2].weight, 2)

        self.assertEqual(weighted[3], 'David')
        self.assertEqual(weighted[3].weight, 4)

        self.assertEqual(weighted[4], 'Eric')
        self.assertEqual(weighted[4].weight, 6)

        self.assertEqual(weighted[5], 'Frank')
        self.assertEqual(weighted[5].weight, 4)

    def test_weight_scale_down(self):
        "Test weight() scales down to max"
        # Add some extras so we can scale them 0.5n+2
        # Weight them so 0=2, 1=2 (rounded down), 4=4, 8=6

        # Eric will be used 8 times total - 6 more
        for i in range(6):
            self.model.objects.create(
                name='Test 3.%d' % i,
                initial_list='Eric',
            )

        # Frank will be used 4 times total - 3 more
        for i in range(3):
            self.model.objects.create(
                name='Test 4.%d' % i,
                initial_list='Frank',
            )

        self.assertTagModel(self.model.initial_list, {
            'Adam':  0,
            'Brian': 0,
            'Chris': 0,
            'David': 1,
            'Eric':  8,
            'Frank': 4,
        })

        weighted = self.tag_model.objects.weight(min=2, max=6)
        self.assertEqual(len(weighted), 6)
        self.assertEqual(weighted[0].name, 'Adam')
        self.assertEqual(weighted[0].weight, 2)

        self.assertEqual(weighted[1], 'Brian')
        self.assertEqual(weighted[1].weight, 2)

        self.assertEqual(weighted[2], 'Chris')
        self.assertEqual(weighted[2].weight, 2)

        self.assertEqual(weighted[3], 'David')
        self.assertEqual(weighted[3].weight, 2)

        self.assertEqual(weighted[4], 'Eric')
        self.assertEqual(weighted[4].weight, 6)

        self.assertEqual(weighted[5], 'Frank')
        self.assertEqual(weighted[5].weight, 4)

    def test_to_string(self):
        "Check manager and queryset can be converted to a tag string"
        self.assertEqual(
            six.text_type(self.tag_model.objects),
            'Adam, Brian, Chris, David, Eric, Frank',
        )
        self.assertEqual(
            six.text_type(self.tag_model.objects.all()),
            'Adam, Brian, Chris, David, Eric, Frank',
        )
        self.assertEqual(
            six.text_type(self.tag_model.objects.initial()),
            'Adam, Brian, Chris',
        )
        self.assertEqual(
            six.text_type(self.o1.initial_list),
            'David, Eric',
        )


###############################################################################
####### Test tag model with custom to_base
###############################################################################

class TagModelToBaseTest(TestCase):
    """
    Test autogenerated tag model with to_base
    """
    manage_models = [
        test_models.TagFieldOptionsModel,
    ]

    def test_custom_base_used(self):
        tag_model = test_models.CustomTagBaseTest.singletag.tag_model
        self.assertTrue(issubclass(tag_model, test_models.CustomTagBase))
        self.assertTrue(issubclass(tag_model, tag_models.TagModel))
        self.assertTrue(tag_model.is_custom)



"""
Test models
"""
from __future__ import unicode_literals

from django.db import models
from django.utils import six

import tagulous


###############################################################################
####### Models for testing TagModel
###############################################################################

class TagMetaAbstractModel(tagulous.models.TagModel):
    """
    An abstract tag model with TagMeta definition
    """
    class Meta:
        abstract = True
    class TagMeta:
        initial = 'Adam, Brian, Chris'
        force_lowercase = True
        max_count = 5

class TagMetaModel(TagMetaAbstractModel):
    """
    A tag model which inherits from TagMetaAbstractModel, with new and changed
    TagMeta values
    """
    class TagMeta:
        max_count = 10
        case_sensitive = True

class TagSlugShorterModel(tagulous.models.BaseTagModel):
    """
    A tag model with a slug field shorter than the name
    """
    name = models.CharField(max_length=20, unique=True)
    slug = models.SlugField(max_length=10)

    # Other fields we're not interested in testing but need to have anyway
    count = models.IntegerField(default=0)
    protected = models.BooleanField(default=False)

    class Meta:
        ordering = ('name',)
        unique_together = (('slug',),)


class TagMetaUser(models.Model):
    """
    A tagged model which uses the TagMetaModel
    """
    name = models.CharField(blank=True, max_length=100)
    two = tagulous.models.TagField(TagMetaModel, blank=True)


###############################################################################
####### Models for testing SingleTagField
###############################################################################

class SingleTagFieldModel(models.Model):
    """
    For testing simple single tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    title = tagulous.models.SingleTagField(blank=True, null=True)

class SingleTagFieldOptionalModel(models.Model):
    """
    Test optional single tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.SingleTagField(blank=True, null=True)

class SingleTagFieldRequiredModel(models.Model):
    """
    Test required single tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.SingleTagField(blank=False, null=False)


class SingleTagFieldMultipleModel(models.Model):
    """
    For testing multiple single tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    tag1 = tagulous.models.SingleTagField(blank=False, null=False)
    tag2 = tagulous.models.SingleTagField(blank=False, null=False)
    tag3 = tagulous.models.SingleTagField(blank=False, null=False)


class SingleTagFieldOptionsModel(models.Model):
    """
    For testing model and form SingleTagField options
    """
    name = models.CharField(blank=True, max_length=100)
    initial_string = tagulous.models.SingleTagField(
        blank=True, null=True, initial='Mr, Mrs, Ms',
    )
    initial_list = tagulous.models.SingleTagField(
        blank=True, null=True, initial=['Mr', 'Mrs', 'Ms'],
        autocomplete_initial=True,
    )
    protect_initial_true = tagulous.models.SingleTagField(
        blank=True, null=True, protect_initial=True, initial='Mr',
    )
    protect_initial_false = tagulous.models.SingleTagField(
        blank=True, null=True, protect_initial=False, initial='Mr',
    )
    protect_all_true = tagulous.models.SingleTagField(
        blank=True, null=True, protect_all=True,
    )
    protect_all_false = tagulous.models.SingleTagField(
        blank=True, null=True, protect_all=False,
    )
    case_sensitive_true = tagulous.models.SingleTagField(
        blank=True, null=True, case_sensitive=True, initial='Mr',
    )
    case_sensitive_false = tagulous.models.SingleTagField(
        blank=True, null=True, case_sensitive=False, initial='Mr',
    )
    force_lowercase_true = tagulous.models.SingleTagField(
        blank=True, null=True, force_lowercase=True,
    )
    force_lowercase_false = tagulous.models.SingleTagField(
        blank=True, null=True, force_lowercase=False,
    )
    # max_count doesn't apply to SingleTagField
    autocomplete_view = tagulous.models.SingleTagField(
        blank=True, null=True,
        autocomplete_view='tagulous_tests_app-null',
    )
    autocomplete_limit = tagulous.models.SingleTagField(
        blank=True, null=True,
        autocomplete_limit=3,
        # Limit only takes effect when there's a view
        autocomplete_view='tagulous_tests_app-null',
    )
    autocomplete_settings = tagulous.models.SingleTagField(
        blank=True, null=True, autocomplete_settings={
            'setting1': 1,
            'setting2': True,
            'setting3': 'example',
        }
    )
    class Meta:
        # Must set a short verbose name - tagulous auto-generated model
        # verbose names will be too long otherwise
        verbose_name = 'STFOM'
        ordering = ('name',)


###############################################################################
####### Models for testing TagField
###############################################################################

class TagFieldModel(models.Model):
    """
    For testing basic tags
    """
    name = models.CharField(blank=True, max_length=100)
    tags = tagulous.models.TagField()

class TagFieldOptionalModel(models.Model):
    """
    Test optional tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.TagField(blank=True)

class TagFieldRequiredModel(models.Model):
    """
    Test required tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    tag = tagulous.models.TagField(blank=False, null=False)

class TagFieldMultipleModel(models.Model):
    """
    For testing multiple tag fields
    """
    name = models.CharField(blank=True, max_length=100)
    tags1 = tagulous.models.TagField(blank=False, null=False)
    tags2 = tagulous.models.TagField(blank=False, null=False)
    tags3 = tagulous.models.TagField(blank=False, null=False)

class TagFieldOptionsModel(models.Model):
    """
    For testing model and form TagField options
    """
    name = models.CharField(blank=True, max_length=100)
    initial_string = tagulous.models.TagField(
        blank=True, initial='Adam, Brian, Chris',
    )
    initial_list = tagulous.models.TagField(
        blank=True, initial=['Adam', 'Brian', 'Chris'],
        autocomplete_initial=True,
    )
    protect_initial_true = tagulous.models.TagField(
        blank=True, protect_initial=True, initial='Adam',
    )
    protect_initial_false = tagulous.models.TagField(
        blank=True, protect_initial=False, initial='Adam',
    )
    protect_all_true = tagulous.models.TagField(
        blank=True, protect_all=True,
    )
    protect_all_false = tagulous.models.TagField(
        blank=True, protect_all=False,
    )
    case_sensitive_true = tagulous.models.TagField(
        blank=True, case_sensitive=True, initial='Adam',
    )
    case_sensitive_false = tagulous.models.TagField(
        blank=True, case_sensitive=False, initial='Adam',
    )
    force_lowercase_true = tagulous.models.TagField(
        blank=True, force_lowercase=True,
    )
    force_lowercase_false = tagulous.models.TagField(
        blank=True, force_lowercase=False,
    )
    # case_sensitive_true_force_lowercase_true - abbreviated to avoid problems
    # with databases that have field name length limits
    cs_true_fl_true = tagulous.models.TagField(
        blank=True, case_sensitive=False, force_lowercase=True,
        verbose_name_singular='case sensitive test',
    )
    max_count = tagulous.models.TagField(
        blank=True, max_count=3,
    )
    autocomplete_view = tagulous.models.TagField(
        blank=True,
        autocomplete_view='tagulous_tests_app-unlimited',
    )
    autocomplete_limit = tagulous.models.TagField(
        blank=True,
        autocomplete_limit=3,
        # Limit only takes effect when there's a view
        autocomplete_view='tagulous_tests_app-limited',
    )
    autocomplete_settings = tagulous.models.TagField(
        blank=True, autocomplete_settings={
            'setting1': 1,
            'setting2': True,
            'setting3': 'example',
        }
    )
    class Meta:
        # Set a short verbose name for tagulous auto-generated verbose name
        verbose_name = 'TFOM'
        ordering = ('name',)



###############################################################################
####### Models for testing a mix of fields
###############################################################################

class SimpleMixedTest(models.Model):
    """
    For tests which need a SingleTagField and TagField
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(blank=True)
    tags = tagulous.models.TagField(blank=True)

class MixedTestTagModel(tagulous.models.TagModel):
    class TagMeta:
        get_absolute_url = lambda self: 'url for %s' % self

class MixedTest(models.Model):
    """
    For tests where it's useful for the SingleTagField and TagField to share
    a tag model
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        MixedTestTagModel, related_name='mixed_singletag',
        blank=True,
    )
    tags = tagulous.models.TagField(
        MixedTestTagModel, related_name='mixed_tags',
        blank=True,
    )
    class Meta:
        ordering = ('name',)

class MixedRefTest(models.Model):
    """
    Multiple models referencing tag tables
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        MixedTest.singletag.tag_model, related_name='mixed_ref_singletag',
        blank=True,
    )
    tags = tagulous.models.TagField(
        MixedTest.tags.tag_model, related_name='mixed_ref_tags',
        blank=True,
    )

class NonTagRefTest(models.Model):
    """
    ForeignKeys and ManyToManyFields directly referencing a tag model
    """
    name = models.CharField(max_length=10)
    fk = models.ForeignKey(
        MixedTest.singletag.tag_model, related_name='non_tag_fk',
        blank=True, on_delete=models.CASCADE,
    )
    mm = models.ManyToManyField(
        MixedTest.tags.tag_model, related_name='non_tag_mm',
        blank=True,
    )


class MixedNonTagModel(tagulous.models.TagModel):
    pass
class MixedNonTagRefTest(models.Model):
    """
    Tag fields and conventional relationships referencing a tag model
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(MixedNonTagModel, blank=True, related_name='singletags')
    tags = tagulous.models.TagField(MixedNonTagModel, blank=True, related_name='tags')
    fk = models.ForeignKey(MixedNonTagModel, blank=True, null=True, related_name='fk', on_delete=models.CASCADE)
    mm = models.ManyToManyField(MixedNonTagModel, blank=True, related_name='mm')


class MixedOrderTest(models.Model):
    """
    For testing ordering of a SingleTagField and TagField when next to other
    M2M and non-M2M fields
    """
    char1   = models.CharField(blank=True, max_length=10)
    fk1     = models.ForeignKey(MixedTest, related_name="order_fk1", on_delete=models.CASCADE)
    char2   = models.CharField(blank=True, max_length=10)
    single1 = tagulous.models.SingleTagField()
    char3   = models.CharField(blank=True, max_length=10)
    m2m1    = models.ManyToManyField(MixedTest, related_name="order_m2m1")
    char4   = models.CharField(blank=True, max_length=10)
    multi1  = tagulous.models.TagField()
    char5   = models.CharField(blank=True, max_length=10)
    m2m2    = models.ManyToManyField(MixedTest, related_name="order_m2m2")
    char6   = models.CharField(blank=True, max_length=10)
    fk2     = models.ForeignKey(MixedTest, related_name="order_fk2", on_delete=models.CASCADE)
    char7   = models.CharField(blank=True, max_length=10)


class MixedStringTagModel(tagulous.models.TagModel):
    pass
class MixedStringTo(models.Model):
    """
    A tagged model with fields which refers to a tag model by string, rather
    than by class
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        'MixedStringTagModel', related_name='tag_meta_string_singletag',
        blank=True,
    )
    tags = tagulous.models.TagField(
        'MixedStringTagModel', related_name='tag_meta_string_tags',
        blank=True,
    )

class MixedSelfTo(tagulous.models.TagModel):
    """
    A tagged tag model, with tag fields which refers itself using 'self'
    """
    alternate = tagulous.models.SingleTagField('self', blank=True)
    related = tagulous.models.TagField('self', blank=True)
    class TagMeta:
        force_lowercase = True



class TreeTest(models.Model):
    """
    For testing tag trees
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(tree=True, blank=True)
    tags = tagulous.models.TagField(tree=True, blank=True)

class CustomTagTree(tagulous.models.TagTreeModel):
    """
    Custom tag tree model
    """
    pass

class CustomTreeTest(models.Model):
    """
    For testing custom tag trees
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        'CustomTagTree', blank=True, related_name='custom_singletag',
    )
    tags = tagulous.models.TagField(
        'CustomTagTree', blank=True, related_name='custom_tags',
    )

class ManyToOneTest(models.Model):
    """
    Add a reverse FK to MixedRefTest for serialization tests
    """
    name = models.CharField(max_length=10)
    mixed_ref_test = models.ForeignKey(MixedRefTest, related_name='many_to_one', on_delete=models.CASCADE)


class CustomTagBase(tagulous.models.TagModel):
    """
    For testing custom tag base models for auto-generated tag models
    """
    is_custom = True

    class Meta:
        abstract = True

class CustomTagBaseTest(models.Model):
    """
    Test custom tag base model
    """
    name = models.CharField(max_length=10)
    singletag = tagulous.models.SingleTagField(
        to_base=CustomTagBase, blank=True, related_name='custom_singletag',
    )
